/* SPDX-License-Identifier: GPL-2.0 */
/*
 * grsecurity - access control and security hardening for Linux
 * Copyright (C) 2001-2019 Bradley Spengler, Open Source Security, Inc.
 * http://www.grsecurity.net spender@grsecurity.net
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef _ASM_POWERPC_ATOMIC_H_
#define _ASM_POWERPC_ATOMIC_H_

/*
 * PowerPC atomic operations
 */

#ifdef __KERNEL__
#include <linux/types.h>
#include <linux/stringify.h>
#include <asm/cmpxchg.h>
#include <asm/barrier.h>
#include <asm/asm-405.h>

#define ATOMIC_INIT(i)		{ (i) }

#define _ASM_EXTABLE(from, to)			\
"	.section	__ex_table,\"a\"\n"	\
"	.balign 4\n"				\
"	.long " #from " - . ;\n"		\
"	.long " #to " - . ;\n"			\
"	.previous\n"

/*
 * Since *_return_relaxed and {cmp}xchg_relaxed are implemented with
 * a "bne-" instruction at the end, so an isync is enough as a acquire barrier
 * on the platform without lwsync.
 */
#define __atomic_acquire_fence()					\
	__asm__ __volatile__(PPC_ACQUIRE_BARRIER "" : : : "memory")

#define __atomic_release_fence()					\
	__asm__ __volatile__(PPC_RELEASE_BARRIER "" : : : "memory")

static __inline__ int atomic_read(const atomic_t *v)
{
	int t;

	__asm__ __volatile__("lwz%U1%X1 %0,%1" : "=r"(t) : "m"(v->counter));

	return t;
}

static __inline__ int atomic_read_unchecked(const atomic_unchecked_t *v)
{
	int t;

	__asm__ __volatile__("lwz%U1%X1 %0,%1" : "=r"(t) : "m"(v->counter));

	return t;
}

static __inline__ void atomic_set(atomic_t *v, int i)
{
	__asm__ __volatile__("stw%U0%X0 %1,%0" : "=m"(v->counter) : "r"(i));
}

static __inline__ void atomic_set_unchecked(atomic_unchecked_t *v, int i)
{
	__asm__ __volatile__("stw%U0%X0 %1,%0" : "=m"(v->counter) : "r"(i));
}

#ifdef CONFIG_PAX_REFCOUNT
#define __REFCOUNT_OP(op) #op"o."
#define __OVERFLOW_PRE			\
	ASM_FTR_IFSET(PPC_MCRXRX_CR0, "	mcrxr cr0\n", %[v3_isa])
#define __OVERFLOW_POST			\
	"	bf+ 4*cr0+so, 3f\n"	\
	"2:	.long 0x00c00b00\n"	\
	"3:\n"
#define __OVERFLOW_EXTABLE		\
	"\n4:\n"			\
	_ASM_EXTABLE(2b, 4b)
#else
#define __REFCOUNT_OP(op) #op
#define __OVERFLOW_PRE
#define __OVERFLOW_POST
#define __OVERFLOW_EXTABLE
#endif

#define __ATOMIC_OP(op, suffix, pre_op, asm_op, post_op, extable)	\
static inline void atomic_##op##suffix(int a, atomic##suffix##_t *v)	\
{									\
	int t;								\
									\
	__asm__ __volatile__(						\
"1:	lwarx	%0,0,%3		# atomic_" #op #suffix "\n"		\
	pre_op								\
	#asm_op " %0,%2,%0\n"						\
	post_op								\
	PPC405_ERR77(0,%3)						\
"	stwcx.	%0,0,%3 \n"						\
"	bne-	1b\n"							\
	extable								\
	: "=&r" (t), "+m" (v->counter)					\
	: "r" (a), "r" (&v->counter), [v3_isa] "i" (CPU_FTR_ARCH_300) 	\
	: "cc", "xer");							\
}									\

#define ATOMIC_OP(op, asm_op) __ATOMIC_OP(op, _unchecked , , asm_op, , )\
			      __ATOMIC_OP(op, , __OVERFLOW_PRE, asm_op##o., __OVERFLOW_POST, __OVERFLOW_EXTABLE)

#define __ATOMIC_OP_RETURN(op, suffix, pre_op, asm_op, post_op, extable)\
static inline int atomic_##op##_return##suffix##_relaxed(int a, atomic##suffix##_t *v)\
{									\
	int t;								\
									\
	__asm__ __volatile__(						\
"1:	lwarx	%0,0,%3		# atomic_" #op "_return" #suffix "_relaxed\n"\
	pre_op								\
	#asm_op " %0,%2,%0\n"						\
	post_op								\
	PPC405_ERR77(0, %3)						\
"	stwcx.	%0,0,%3\n"						\
"	bne-	1b\n"							\
	extable								\
	: "=&r" (t), "+m" (v->counter)					\
	: "r" (a), "r" (&v->counter), [v3_isa] "i" (CPU_FTR_ARCH_300)	\
	: "cc", "xer");							\
									\
	return t;							\
}

#define ATOMIC_OP_RETURN_RELAXED(op, asm_op) __ATOMIC_OP_RETURN(op, _unchecked, , asm_op, , )\
				     __ATOMIC_OP_RETURN(op, , __OVERFLOW_PRE, asm_op##o., __OVERFLOW_POST, __OVERFLOW_EXTABLE)

#define __ATOMIC_FETCH_OP_RELAXED(op, suffix, pre_op, asm_op, post_op, extable) \
static inline int atomic_fetch_##op##suffix##_relaxed(int a, atomic##suffix##_t *v) \
{									\
	int res, t;							\
									\
	__asm__ __volatile__(						\
"1:	lwarx	%0,0,%4		# atomic_fetch_" #op #suffix "_relaxed\n"\
	pre_op								\
	#asm_op " %1,%3,%0\n"						\
	post_op								\
	PPC405_ERR77(0, %4)						\
"	stwcx.	%1,0,%4\n"						\
"	bne-	1b\n"							\
	extable								\
	: "=&r" (res), "=&r" (t), "+m" (v->counter)			\
	: "r" (a), "r" (&v->counter), [v3_isa] "i" (CPU_FTR_ARCH_300)	\
	: "cc", "xer");							\
									\
	return res;							\
}

#define ATOMIC_FETCH_OP_RELAXED(op, asm_op) __ATOMIC_FETCH_OP_RELAXED(op, _unchecked, , asm_op, , ) \
				    __ATOMIC_FETCH_OP_RELAXED(op, , __OVERFLOW_PRE, asm_op##o., __OVERFLOW_POST, __OVERFLOW_EXTABLE)

#define ATOMIC_OPS(op, asm_op)						\
	ATOMIC_OP(op, asm_op)						\
	ATOMIC_OP_RETURN_RELAXED(op, asm_op)				\
	ATOMIC_FETCH_OP_RELAXED(op, asm_op)

ATOMIC_OPS(add, add)
ATOMIC_OPS(sub, subf)

#define atomic_add_return_relaxed atomic_add_return_relaxed
#define atomic_add_return_unchecked_relaxed atomic_add_return_unchecked_relaxed
#define atomic_sub_return_relaxed atomic_sub_return_relaxed

#define atomic_fetch_add_relaxed atomic_fetch_add_relaxed
#define atomic_fetch_sub_relaxed atomic_fetch_sub_relaxed

#undef ATOMIC_OPS
#undef ATOMIC_OP
#undef ATOMIC_FETCH_OP_RELAXED

#define ATOMIC_OP(op, asm_op) __ATOMIC_OP(op, , , asm_op, , )
#define ATOMIC_FETCH_OP_RELAXED(op, asm_op) __ATOMIC_FETCH_OP_RELAXED(op, , , asm_op, , )

#define ATOMIC_OPS(op, asm_op)						\
	ATOMIC_OP(op, asm_op)						\
	ATOMIC_FETCH_OP_RELAXED(op, asm_op)

ATOMIC_OPS(and, and)
ATOMIC_OPS(or, or)
ATOMIC_OPS(xor, xor)

#define atomic_fetch_and_relaxed atomic_fetch_and_relaxed
#define atomic_fetch_or_relaxed  atomic_fetch_or_relaxed
#define atomic_fetch_xor_relaxed atomic_fetch_xor_relaxed

#undef ATOMIC_OPS
#undef ATOMIC_FETCH_OP_RELAXED
#undef ATOMIC_OP_RETURN_RELAXED
#undef __ATOMIC_OP_RETURN
#undef ATOMIC_OP
#undef __ATOMIC_OP

/*
 * atomic_inc - increment atomic variable
 * @v: pointer of type atomic_t
 *
 * Automatically increments @v by 1
 */
#define atomic_inc(v) atomic_add(1, (v))
#define atomic_inc_unchecked(v) atomic_add_unchecked(1, (v))
#define atomic_inc_return_relaxed(v) atomic_add_return_relaxed(1, (v))
#define atomic_inc_return_unchecked_relaxed(v) atomic_add_return_unchecked_relaxed(1, (v))

#define atomic_dec_return_relaxed(v) atomic_sub_return_relaxed(1, (v))

#define atomic_cmpxchg(v, o, n) (cmpxchg(&((v)->counter), (o), (n)))
#define atomic_cmpxchg_relaxed(v, o, n) \
	cmpxchg_relaxed(&((v)->counter), (o), (n))
#define atomic_cmpxchg_acquire(v, o, n) \
	cmpxchg_acquire(&((v)->counter), (o), (n))

#define atomic_xchg(v, new) (xchg(&((v)->counter), new))
#define atomic_xchg_relaxed(v, new) xchg_relaxed(&((v)->counter), (new))

#define atomic_cmpxchg_unchecked(v, o, n) (cmpxchg(&((v)->counter), (o), (n)))
#define atomic_xchg_unchecked(v, new) (xchg(&((v)->counter), new))

/**
 * atomic_fetch_add_unless - add unless the number is a given value
 * @v: pointer of type atomic_t
 * @a: the amount to add to v...
 * @u: ...unless v is equal to u.
 *
 * Atomically adds @a to @v, so long as it was not @u.
 * Returns the old value of @v.
 */
static __inline__ int atomic_fetch_add_unless(atomic_t *v, int a, int u)
{
	int t;

	__asm__ __volatile__ (
	PPC_ATOMIC_ENTRY_BARRIER
"1:	lwarx	%0,0,%1		# atomic_fetch_add_unless\n\
	cmpw	0,%0,%3 \n\
	beq	5f \n"

	__OVERFLOW_PRE
	__REFCOUNT_OP(add) "	%0,%2,%0 \n"
	__OVERFLOW_POST

	PPC405_ERR77(0,%2)
"	stwcx.	%0,0,%1 \n\
	bne-	1b \n"

	__OVERFLOW_EXTABLE

	PPC_ATOMIC_EXIT_BARRIER
"	subf	%0,%2,%0 \n\
5:"
	: "=&r" (t)
	: "r" (&v->counter), "r" (a), "r" (u), [v3_isa] "i" (CPU_FTR_ARCH_300)
	: "cc", "xer", "memory");

	return t;
}
#define atomic_fetch_add_unless atomic_fetch_add_unless

/*
 * Atomically test *v and decrement if it is greater than 0.
 * The function returns the old value of *v minus 1, even if
 * the atomic variable, v, was not decremented.
 */
static __inline__ int atomic_dec_if_positive(atomic_t *v)
{
	int t;

	__asm__ __volatile__(
	PPC_ATOMIC_ENTRY_BARRIER
"1:	lwarx	%0,0,%1		# atomic_dec_if_positive\n\
	cmpwi	%0,1\n\
	addi	%0,%0,-1\n\
	blt-	2f\n"
	PPC405_ERR77(0,%1)
"	stwcx.	%0,0,%1\n\
	bne-	1b"
	PPC_ATOMIC_EXIT_BARRIER
	"\n\
2:"	: "=&b" (t)
	: "r" (&v->counter)
	: "cc", "memory");

	return t;
}
#define atomic_dec_if_positive atomic_dec_if_positive

#ifdef __powerpc64__

#define ATOMIC64_INIT(i)	{ (i) }

static __inline__ s64 atomic64_read(const atomic64_t *v)
{
	s64 t;

	__asm__ __volatile__("ld%U1%X1 %0,%1" : "=r"(t) : "m"(v->counter));

	return t;
}

static __inline__ s64 atomic64_read_unchecked(const atomic64_unchecked_t *v)
{
	s64 t;

	__asm__ __volatile__("ld%U1%X1 %0,%1" : "=r"(t) : "m"(v->counter));

	return t;
}

static __inline__ void atomic64_set(atomic64_t *v, s64 i)
{
	__asm__ __volatile__("std%U0%X0 %1,%0" : "=m"(v->counter) : "r"(i));
}

static __inline__ void atomic64_set_unchecked(atomic64_unchecked_t *v, s64 i)
{
	__asm__ __volatile__("std%U0%X0 %1,%0" : "=m"(v->counter) : "r"(i));
}

#define __ATOMIC64_OP(op, suffix, pre_op, asm_op, post_op, extable)	\
static inline void atomic64_##op##suffix(s64 a, atomic64##suffix##_t *v)\
{									\
	s64 t;								\
									\
	__asm__ __volatile__(						\
"1:	ldarx	%0,0,%3		# atomic64_" #op "\n"			\
	pre_op								\
	#asm_op " %0,%2,%0\n"						\
	post_op								\
"	stdcx.	%0,0,%3 \n"						\
"	bne-	1b\n"							\
	extable								\
	: "=&r" (t), "+m" (v->counter)					\
	: "r" (a), "r" (&v->counter), [v3_isa] "i" (CPU_FTR_ARCH_300)	\
	: "cc", "xer");							\
}

#define ATOMIC64_OP(op, asm_op) __ATOMIC64_OP(op, _unchecked, , asm_op, , )		\
				__ATOMIC64_OP(op, , __OVERFLOW_PRE, asm_op##o., __OVERFLOW_POST, __OVERFLOW_EXTABLE)

#define __ATOMIC64_OP_RETURN(op, suffix, pre_op, asm_op, post_op, extable)\
static inline s64 atomic64_##op##_return##suffix##_relaxed(s64 a, atomic64##suffix##_t *v)\
{									\
	s64 t;								\
									\
	__asm__ __volatile__(						\
"1:	ldarx	%0,0,%3		# atomic64_" #op "_return_relaxed\n"	\
	pre_op								\
	#asm_op " %0,%2,%0\n"						\
	post_op								\
"	stdcx.	%0,0,%3\n"						\
"	bne-	1b\n"							\
	extable								\
	: "=&r" (t), "+m" (v->counter)					\
	: "r" (a), "r" (&v->counter), [v3_isa] "i" (CPU_FTR_ARCH_300)	\
	: "cc", "xer");							\
									\
	return t;							\
}

#define ATOMIC64_OP_RETURN_RELAXED(op, asm_op) __ATOMIC64_OP_RETURN(op, _unchecked, , asm_op, , )\
				       __ATOMIC64_OP_RETURN(op, , __OVERFLOW_PRE, asm_op##o., __OVERFLOW_POST, __OVERFLOW_EXTABLE)

#define __ATOMIC64_FETCH_OP_RELAXED(op, suffix, pre_op, asm_op, post_op, extable) \
static inline s64							\
atomic64_fetch_##op##suffix##_relaxed(s64 a, atomic64##suffix##_t *v)	\
{									\
	s64 res, t;							\
									\
	__asm__ __volatile__(						\
"1:	ldarx	%0,0,%4		# atomic64_fetch_" #op #suffix "_relaxed\n"\
	pre_op								\
	#asm_op " %1,%3,%0\n"						\
	post_op								\
"	stdcx.	%1,0,%4\n"						\
"	bne-	1b\n"							\
	extable								\
	: "=&r" (res), "=&r" (t), "+m" (v->counter)			\
	: "r" (a), "r" (&v->counter), [v3_isa] "i" (CPU_FTR_ARCH_300)	\
	: "cc", "xer");							\
									\
	return res;							\
}

#define ATOMIC64_FETCH_OP_RELAXED(op, asm_op) __ATOMIC64_FETCH_OP_RELAXED(op, _unchecked, , asm_op, , ) \
				    __ATOMIC64_FETCH_OP_RELAXED(op, , __OVERFLOW_PRE, asm_op##o., __OVERFLOW_POST, __OVERFLOW_EXTABLE)

#define ATOMIC64_OPS(op, asm_op)					\
	ATOMIC64_OP(op, asm_op)						\
	ATOMIC64_OP_RETURN_RELAXED(op, asm_op)				\
	ATOMIC64_FETCH_OP_RELAXED(op, asm_op)

ATOMIC64_OPS(add, add)
ATOMIC64_OPS(sub, subf)

#define atomic64_add_return_relaxed atomic64_add_return_relaxed
#define atomic64_add_return_unchecked_relaxed atomic64_add_return_unchecked_relaxed
#define atomic64_sub_return_relaxed atomic64_sub_return_relaxed

#define atomic64_fetch_add_relaxed atomic64_fetch_add_relaxed
#define atomic64_fetch_sub_relaxed atomic64_fetch_sub_relaxed

#undef ATOMIC64_OPS
#undef ATOMIC64_OP
#undef ATOMIC64_FETCH_OP_RELAXED

#define ATOMIC64_OP(op, asm_op) __ATOMIC64_OP(op, , , asm_op, , )
#define ATOMIC64_FETCH_OP_RELAXED(op, asm_op) __ATOMIC64_FETCH_OP_RELAXED(op, , , asm_op, , )

#define ATOMIC64_OPS(op, asm_op)					\
	ATOMIC64_OP(op, asm_op)						\
	ATOMIC64_FETCH_OP_RELAXED(op, asm_op)

ATOMIC64_OPS(and, and)
ATOMIC64_OPS(or, or)
ATOMIC64_OPS(xor, xor)

#define atomic64_fetch_and_relaxed atomic64_fetch_and_relaxed
#define atomic64_fetch_or_relaxed  atomic64_fetch_or_relaxed
#define atomic64_fetch_xor_relaxed atomic64_fetch_xor_relaxed

#undef ATOPIC64_OPS
#undef ATOMIC64_FETCH_OP_RELAXED
#undef ATOMIC64_OP_RETURN_RELAXED
#undef __ATOMIC64_OP_RETURN
#undef ATOMIC64_OP
#undef __ATOMIC64_OP

/*
 * atomic64_inc - increment atomic variable
 * @v: pointer of type atomic64_t
 *
 * Automatically increments @v by 1
 */
#define atomic64_inc(v) atomic64_add(1, (v))
#define atomic64_inc_return_relaxed(v) atomic64_add_return_relaxed(1, (v))

static inline s64 atomic64_inc_return_unchecked_relaxed(atomic64_unchecked_t *v)
{
	return atomic64_add_return_unchecked_relaxed(1, v);
}
#define atomic64_inc_return_unchecked_relaxed atomic64_inc_return_unchecked_relaxed

/* 
 * atomic64_dec - decrement atomic variable
 * @v: pointer of type atomic64_t
 * 
 * Atomically decrements @v by 1
 */
#define atomic64_dec(v) atomic64_sub(1, (v))
#define atomic64_dec_return_relaxed(v) atomic64_sub_return_relaxed(1, (v))

/*
 * Atomically test *v and decrement if it is greater than 0.
 * The function returns the old value of *v minus 1.
 */
static __inline__ s64 atomic64_dec_if_positive(atomic64_t *v)
{
	s64 t;

	__asm__ __volatile__(
	PPC_ATOMIC_ENTRY_BARRIER
"1:	ldarx	%0,0,%1		# atomic64_dec_if_positive\n\
	addic.	%0,%0,-1\n\
	blt-	2f\n\
	stdcx.	%0,0,%1\n\
	bne-	1b"
	PPC_ATOMIC_EXIT_BARRIER
	"\n\
2:"	: "=&r" (t)
	: "r" (&v->counter)
	: "cc", "xer", "memory");

	return t;
}
#define atomic64_dec_if_positive atomic64_dec_if_positive

#define atomic64_cmpxchg(v, o, n) (cmpxchg(&((v)->counter), (o), (n)))
#define atomic64_cmpxchg_relaxed(v, o, n) \
	cmpxchg_relaxed(&((v)->counter), (o), (n))
#define atomic64_cmpxchg_acquire(v, o, n) \
	cmpxchg_acquire(&((v)->counter), (o), (n))

#define atomic64_xchg(v, new) (xchg(&((v)->counter), new))
#define atomic64_xchg_relaxed(v, new) xchg_relaxed(&((v)->counter), (new))

#define atomic64_cmpxchg_unchecked(v, o, n)	atomic64_cmpxchg((v), (o), (n))
#define atomic64_xchg_unchecked(v, n)		atomic64_xchg((v), (n))

/**
 * atomic64_fetch_add_unless - add unless the number is a given value
 * @v: pointer of type atomic64_t
 * @a: the amount to add to v...
 * @u: ...unless v is equal to u.
 *
 * Atomically adds @a to @v, so long as it was not @u.
 * Returns the old value of @v.
 */
static __inline__ s64 atomic64_fetch_add_unless(atomic64_t *v, s64 a, s64 u)
{
	s64 t;

	__asm__ __volatile__ (
	PPC_ATOMIC_ENTRY_BARRIER
"1:	ldarx	%0,0,%1		# atomic64_fetch_add_unless\n\
	cmpd	0,%0,%3 \n\
	beq	5f \n"

	__OVERFLOW_PRE
	__REFCOUNT_OP(add) "	%0,%2,%0 \n"
	__OVERFLOW_POST

"	stdcx.	%0,0,%1 \n\
	bne-	1b \n"

	__OVERFLOW_EXTABLE

	PPC_ATOMIC_EXIT_BARRIER

"	subf	%0,%2,%0 \n\
5:"
	: "=&r" (t)
	: "r" (&v->counter), "r" (a), "r" (u), [v3_isa] "i" (CPU_FTR_ARCH_300)
	: "cc", "xer", "memory");

	return t;
}
#define atomic64_fetch_add_unless atomic64_fetch_add_unless

#define atomic64_inc_not_zero(v)	atomic64_add_unless((v), 1, 0)

#undef __OVERFLOW_EXTABLE
#undef __OVERFLOW_POST
#undef __OVERFLOW_PRE
#undef __REFCOUNT_OP

#endif /* __powerpc64__ */

#endif /* __KERNEL__ */
#endif /* _ASM_POWERPC_ATOMIC_H_ */
