/*
 * Copyright 2012-2020 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * Note: the choice of the license means that the compilation process is
 *       NOT 'eligible' as defined by gcc's library exception to the GPL v3,
 *       but for the kernel it doesn't matter since it doesn't link against
 *       any of the gcc libraries
 *
 * gcc plugin to provide various features for older gcc versions or other plugins
 * <gcc-4.9: colorize diagnostic output
 * <gcc-8:   emergency dump on ICE
 * plugins:  context tracking infrastructure
 *           symbolic kernel constants via ksyms.def
 *
 */

#include "gcc-common.h"
#include "gcc-ksyms.h"

__visible int plugin_is_GPL_compatible;

static struct plugin_info utilities_plugin_info = {
	.version	= "202004061642",
	.help		= "color=[never|always|auto]\tdetermine when to colorize\n"
			  "include=<file>\t\tinclude <file> into all translation units to provide various asm macros\n"
			  "align-text-bits=<bits>\t\talign .init functions to 2^<bits>\n"
};

static const char *asm_include;

#if BUILDING_GCC_VERSION < 4009
#define GREEN		"\033[32m\033[K"
#define LIGHTGREEN	"\033[1;32m\033[K"
#define YELLOW		"\033[33m\033[K"
#define LIGHTYELLOW	"\033[1;33m\033[K"
#define RED		"\033[31m\033[K"
#define LIGHTRED	"\033[1;31m\033[K"
#define BLUE		"\033[34m\033[K"
#define LIGHTBLUE	"\033[1;34m\033[K"
#define BRIGHT		"\033[1;m\033[K"
#define NORMAL		"\033[m\033[K"

static diagnostic_starter_fn old_starter;
static diagnostic_finalizer_fn old_finalizer;

static void start_colorize(diagnostic_context *context, diagnostic_info *diagnostic)
{
	const char *color;
	char *newprefix;

	switch (diagnostic->kind) {
	case DK_NOTE:
		color = LIGHTBLUE;
		break;

	case DK_PEDWARN:
	case DK_WARNING:
		color = LIGHTYELLOW;
		break;

	case DK_ERROR:
	case DK_FATAL:
	case DK_ICE:
	case DK_PERMERROR:
	case DK_SORRY:
		color = LIGHTRED;
		break;

	default:
		color = NORMAL;
	}

	old_starter(context, diagnostic);
	if (-1 == asprintf(&newprefix, "%s%s" NORMAL, color, context->printer->prefix))
		return;
	pp_destroy_prefix(context->printer);
	pp_set_prefix(context->printer, newprefix);
}

static void finalize_colorize(diagnostic_context *context, diagnostic_info *diagnostic)
{
	old_finalizer(context, diagnostic);
}

static void colorize_arm(void)
{
	old_starter = diagnostic_starter(global_dc);
	old_finalizer = diagnostic_finalizer(global_dc);

	diagnostic_starter(global_dc) = start_colorize;
	diagnostic_finalizer(global_dc) = finalize_colorize;
}

static unsigned int colorize_rearm_execute(void)
{
	if (diagnostic_starter(global_dc) == start_colorize)
		return 0;

	colorize_arm();
	return 0;
}

#define PASS_NAME colorize_rearm
#define NO_GATE
#include "gcc-generate-simple_ipa-pass.h"

static void colorize_start_unit(void *gcc_data __unused, void *user_data __unused)
{
	colorize_arm();
}

static bool should_colorize(void)
{
	char const *t = getenv("TERM");

	return t && strcmp(t, "dumb") && isatty(STDERR_FILENO);
}
#endif

#if BUILDING_GCC_VERSION < 8000
static void execute_function_dump(struct function *fn, void *data ATTRIBUTE_UNUSED)
{
	if (!dump_file)
		return;

	push_cfun(fn);

	if (fn->curr_properties & PROP_trees)
		dump_function_to_file(fn->decl, dump_file, dump_flags);
	else
		print_rtl_with_bb(dump_file, get_insns(), dump_flags);

	fflush(dump_file);

	if ((cfun->curr_properties & PROP_cfg) && (dump_flags & TDF_GRAPH))
#if BUILDING_GCC_VERSION < 4008
		if (!(fn->curr_properties & PROP_trees))
			print_rtl_graph_with_bb(dump_file_name, get_insns());
#else
		print_graph_cfg(dump_file_name, cfun);
#endif

	pop_cfun();
}

#if BUILDING_GCC_VERSION < 4006
static void (*old_internal_error)(const char *, va_list *);

static void emergency_dump_internal_error_function(const char *msgid, va_list *ap)
#else
static void (*old_internal_error)(diagnostic_context *, const char *, va_list *);

static void emergency_dump_internal_error_function(diagnostic_context *dc, const char *msgid, va_list *ap)
#endif
{
	enum opt_pass_type pt;

	if (old_internal_error)
#if BUILDING_GCC_VERSION < 4006
		old_internal_error(msgid, ap);
#else
		old_internal_error(dc, msgid, ap);
#endif

	if (!current_pass || !dump_file || !cfun)
		return;

	pt = current_pass->type;
	fnotice(stderr, "during %s pass: %s\n", pt == GIMPLE_PASS ? "GIMPLE" : pt == RTL_PASS ? "RTL" : "IPA", current_pass->name);
	fnotice(stderr, "dump file: %s\n", dump_file_name);
	fprintf(dump_file, "\n\n\nEMERGENCY DUMP:\n\n");
	execute_function_dump(cfun, current_pass);
}
#endif

static const struct gcc_debug_hooks *old_debug_hooks;
static struct gcc_debug_hooks utilities_debug_hooks;

static void utilities_assembly_start(void)
{
	// chain to previous callback
	if (old_debug_hooks && old_debug_hooks->assembly_start)
		old_debug_hooks->assembly_start();

	fprintf(asm_out_file, "\t.include \"%s\"\n", asm_include);
}

static void (*old_finish)(void);
static void utilities_finish(void)
{
	deps_add_dep(cpp_get_deps(parse_in), asm_include);
	old_finish();
}

static void include_start_unit(void __unused *gcc_data, void __unused *user_data)
{
	if (debug_hooks)
		utilities_debug_hooks = *debug_hooks;
	utilities_debug_hooks.assembly_start = utilities_assembly_start;

	old_debug_hooks = debug_hooks;
	debug_hooks = &utilities_debug_hooks;

	old_finish = lang_hooks.finish;
	lang_hooks.finish = utilities_finish;
}

static bool split_context_attribute(tree args, tree *lock, tree *in, tree *out)
{
	*in = TREE_VALUE(args);

	if (TREE_CODE(*in) != INTEGER_CST) {
		*lock = *in;
		args = TREE_CHAIN(args);
		*in = TREE_VALUE(args);
	} else
		*lock = NULL_TREE;

	args = TREE_CHAIN(args);
	if (*lock && !args)
		return false;

	*out = TREE_VALUE(args);
	return true;
}

static tree handle_context_attribute(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	*no_add_attrs = true;
	tree lock, in, out;

	if (TREE_CODE(*node) != FUNCTION_DECL) {
		error_at(DECL_P(*node) ? DECL_SOURCE_LOCATION(*node) : UNKNOWN_LOCATION, "%qE attribute applies to functions only (%qD)", name, *node);
		return NULL_TREE;
	}

	if (!split_context_attribute(args, &lock, &in, &out)) {
		error_at(DECL_SOURCE_LOCATION(*node), "%qE attribute needs two integers after the lock expression", name);
		return NULL_TREE;
	}

	if (TREE_CODE(in) != INTEGER_CST) {
		error_at(DECL_SOURCE_LOCATION(*node), "the 'in' argument of the %qE attribute must be an integer (%qE)", name, in);
		return NULL_TREE;
	}

	if (TREE_CODE(out) != INTEGER_CST) {
		error_at(DECL_SOURCE_LOCATION(*node), "the 'out' argument of the %qE attribute must be an integer (%qE)", name, out);
		return NULL_TREE;
	}

	*no_add_attrs = false;
	return NULL_TREE;
}

static tree handle_once_attribute(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	tree *type, orig_type, decl;

	decl = *node;
	gcc_assert(DECL_P(decl));
	gcc_assert(TREE_CODE(decl) == TYPE_DECL);
	gcc_assert(DECL_ORIGINAL_TYPE(decl) == NULL_TREE);
	gcc_assert(!(flags & (int)ATTR_FLAG_TYPE_IN_PLACE));

	type = &TREE_TYPE(decl);
	orig_type = TREE_TYPE(decl);
	if (lookup_attribute("once", TYPE_ATTRIBUTES(orig_type)))
		return NULL_TREE;

	if (RECORD_OR_UNION_TYPE_P(orig_type)) {
		gcc_assert(TYPE_FIELDS(orig_type));
		if (TREE_CHAIN(TYPE_FIELDS(orig_type))) {
//			inform(DECL_SOURCE_LOCATION(decl), "invalid ONCE access to %qD of type %qT: more than one field", decl, orig_type);
			return NULL_TREE;
		}
		if (tree_int_cst_lt(TYPE_SIZE_UNIT(long_unsigned_type_node), TYPE_SIZE_UNIT(orig_type))) {
//			inform(DECL_SOURCE_LOCATION(decl), "invalid ONCE access to %qD of type %qT: size too big", decl, orig_type);
			return NULL_TREE;
		}
	}

	*type = build_duplicate_type(orig_type);
	TYPE_ATTRIBUTES(*type) = tree_cons(name, NULL_TREE, TYPE_ATTRIBUTES(orig_type));
	*no_add_attrs = false;
	return NULL_TREE;
}

static struct attribute_spec context_attr = ATTRIBUTE_SPEC_INIT("context", 2, 3, true, false, false, handle_context_attribute, true, NULL);
static struct attribute_spec once_attr = ATTRIBUTE_SPEC_INIT("once", 0, 0, true, false, false, handle_once_attribute, false, NULL);

static void register_attributes(void *event_data, void *data)
{
	register_attribute(&context_attr);
	register_attribute(&once_attr);
}

static const char context_function[] = "__context__";
static GTY(()) tree context_function_decl;

static const char context_error[] = "__context_error__";
static GTY(()) tree context_error_decl;

static void context_start_unit(void __unused *gcc_data, void __unused *user_data)
{
	tree fntype, attr;

	// void __context__(void *, int);
	fntype = build_function_type_list(void_type_node, ptr_type_node, integer_type_node, NULL_TREE);
	context_function_decl = build_fn_decl(context_function, fntype);

	TREE_PUBLIC(context_function_decl) = 1;
	TREE_USED(context_function_decl) = 1;
	DECL_EXTERNAL(context_function_decl) = 1;
	DECL_ARTIFICIAL(context_function_decl) = 1;
	DECL_PRESERVE_P(context_function_decl) = 1;
//	TREE_NOTHROW(context_function_decl) = 1;
//	DECL_UNINLINABLE(context_function_decl) = 1;
	DECL_ASSEMBLER_NAME(context_function_decl); // for LTO
	lang_hooks.decls.pushdecl(context_function_decl);

	invoke_plugin_callbacks(PLUGIN_PRE_GENERICIZE, context_function_decl);

	// void __context_error__(const void *, int) __attribute__((error("context error")));
	fntype = build_function_type_list(void_type_node, const_ptr_type_node, integer_type_node, NULL_TREE);
	context_error_decl = build_fn_decl(context_error, fntype);

	TREE_PUBLIC(context_error_decl) = 1;
	TREE_USED(context_error_decl) = 1;
	DECL_EXTERNAL(context_error_decl) = 1;
	DECL_ARTIFICIAL(context_error_decl) = 1;
	DECL_PRESERVE_P(context_error_decl) = 1;
//	TREE_NOTHROW(context_error_decl) = 1;
//	DECL_UNINLINABLE(context_error_decl) = 1;
	TREE_THIS_VOLATILE(context_error_decl) = 1;
	DECL_ASSEMBLER_NAME(context_error_decl);

	attr = tree_cons(NULL, build_const_char_string(14, "context error"), NULL);
	attr = tree_cons(get_identifier("error"), attr, NULL);
	decl_attributes(&context_error_decl, attr, 0);

	invoke_plugin_callbacks(PLUGIN_PRE_GENERICIZE, context_error_decl);
}

static basic_block verify_context_before(gimple_stmt_iterator *gsi, tree context, tree inout, tree error)
{
	gimple stmt;
	basic_block cond_bb, join_bb, true_bb;
	edge e;
	cgraph_node_ptr node;
	location_t loc;
	const char *file;
	int frequency, line;
	size_t len;
	tree filename, block;

	stmt = gsi_stmt(*gsi);
	cond_bb = gsi_bb(*gsi);

	block = gimple_block(stmt);
	if (!block) {
		gcc_assert(!bb_loop_depth(cond_bb));
		block = DECL_INITIAL(current_function_decl);
	}

	if (gimple_has_location(stmt)) {
		loc = gimple_location(stmt);
		file = gimple_filename(stmt);
		line = gimple_lineno(stmt);
	} else {
		loc = DECL_SOURCE_LOCATION(current_function_decl);
		file = DECL_SOURCE_FILE(current_function_decl);
		line = DECL_SOURCE_LINE(current_function_decl);
	}
	gcc_assert(file);

	// if (context != count) __context_error__(__FILE__, __LINE__);
	stmt = gimple_build_cond(NE_EXPR, context, inout, NULL_TREE, NULL_TREE);
	gimple_set_location(stmt, loc);
	gimple_set_block(stmt, block);
	gsi_insert_before(gsi, stmt, GSI_NEW_STMT);

	gcc_assert(!gsi_end_p(*gsi));
	gcc_assert(stmt == gsi_stmt(*gsi));

#if 1
	e = split_block(cond_bb, gsi_stmt(*gsi));
	gcc_assert(cond_bb == e->src);
	join_bb = e->dest;
	e->flags = EDGE_FALSE_VALUE;
	e->probability = REG_BR_PROB_EVEN;

	true_bb = create_empty_bb(EXIT_BLOCK_PTR_FOR_FN(cfun)->prev_bb);
	e = make_edge(cond_bb, true_bb, EDGE_TRUE_VALUE);
	e->probability = REG_BR_PROB_EVEN;
	true_bb->count = edge_count(e);

	e = make_single_succ_edge(true_bb, join_bb, EDGE_FALLTHRU);

	gcc_assert(dom_info_available_p(CDI_DOMINATORS));
	set_immediate_dominator(CDI_DOMINATORS, true_bb, cond_bb);
	set_immediate_dominator(CDI_DOMINATORS, join_bb, cond_bb);

	gcc_assert(cond_bb->loop_father == join_bb->loop_father);
	add_bb_to_loop(true_bb, cond_bb->loop_father);
#else
	true_bb = insert_cond_bb(cond_bb, gsi_stmt(*gsi), stmt, REG_BR_PROB_EVEN);
#endif
	loops_state_set(LOOPS_NEED_FIXUP);

	// insert call to builtin_trap or __context_error__
	*gsi = gsi_start_bb(true_bb);

#if 0
	error = builtin_decl_implicit(BUILT_IN_TRAP);
	stmt = gimple_build_call(error, 0);
#else
	len = strlen(file) + 1;
	filename = build_const_char_string(len, file);
	filename = build_fold_addr_expr_with_type_loc(loc, filename, const_ptr_type_node);
	stmt = gimple_build_call(error, 2, filename, build_int_cst(NULL_TREE, line));
#endif
	gimple_set_location(stmt, loc);
	gimple_set_block(stmt, block);
	gsi_insert_after(gsi, stmt, GSI_CONTINUE_LINKING);

	// update the cgraph
	node = cgraph_get_create_node(error);
	gcc_assert(node);
	frequency = compute_call_stmt_bb_frequency(current_function_decl, true_bb);
	cgraph_create_edge(cgraph_get_node(current_function_decl), node, as_a_gcall(stmt), true_bb->count, frequency, true_bb->loop_depth);

	*gsi = gsi_start_nondebug_bb(join_bb);
	return join_bb;
}

static void update_context(gimple_stmt_iterator *gsi, tree context, int diff)
{
	gimple assign;
	tree op;
	location_t loc = gimple_location(gsi_stmt(*gsi));

	op = fold_build2_loc(loc, PLUS_EXPR, integer_type_node, context, build_int_cst(integer_type_node, diff));
	assign = gimple_build_assign(context, op);
	gimple_set_location(assign, loc);
	gimple_set_block(assign, gimple_block(gsi_stmt(*gsi)));
	gsi_insert_after(gsi, assign, GSI_NEW_STMT);
	update_stmt(assign);
}

static basic_block track_context(basic_block bb, tree context)
{
	gimple_stmt_iterator gsi;

	// adjust context according to the context information on any call stmt
	for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
		gimple stmt = gsi_stmt(gsi);
		tree fndecl, context_attr;
		tree lock, in, out;
		int incount, outcount;

		if (!is_gimple_call(stmt))
			continue;

		fndecl = gimple_call_fndecl(stmt);
		if (!fndecl)
			continue;

		if (fndecl == context_function_decl) {
			unsigned int num_ops = gimple_num_ops(stmt);
			int diff = tree_to_shwi(gimple_op(stmt, num_ops - 1));

			gcc_assert(diff);
			update_context(&gsi, context, diff);
			gsi_prev(&gsi);
			gsi_remove(&gsi, true);
			continue;
		}

		context_attr = lookup_attribute("context", DECL_ATTRIBUTES(fndecl));
		if (!context_attr)
			continue;

		gcc_assert(split_context_attribute(TREE_VALUE(context_attr), &lock, &in, &out));
		incount = tree_to_shwi(in);
		outcount = tree_to_shwi(out);
		bb = verify_context_before(&gsi, context, in, context_error_decl);
		update_context(&gsi, context, outcount - incount);
	}

	return bb;
}

static unsigned int context_execute(void)
{
	basic_block bb;
	gimple assign;
	gimple_stmt_iterator gsi;
	tree context_attr, context;
	tree lock, in, out;

	loop_optimizer_init(LOOPS_NORMAL | LOOPS_HAVE_RECORDED_EXITS);
	gcc_assert(current_loops);

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);

	context_attr = lookup_attribute("context", DECL_ATTRIBUTES(current_function_decl));
	if (context_attr) {
		gcc_assert(split_context_attribute(TREE_VALUE(context_attr), &lock, &in, &out));
	} else {
		lock = NULL_TREE;
		in = out = integer_zero_node;
	}

	// 1. create local context variable
	context = create_tmp_var(integer_type_node, "context");
	add_referenced_var(context);
	mark_sym_for_renaming(context);

	// 2. initialize local context variable
	gcc_assert(single_succ_p(ENTRY_BLOCK_PTR_FOR_FN(cfun)));
	bb = single_succ(ENTRY_BLOCK_PTR_FOR_FN(cfun));
	if (!single_pred_p(bb)) {
		gcc_assert(bb_any_loop(bb));
		plugin_split_edge(single_succ_edge(ENTRY_BLOCK_PTR_FOR_FN(cfun)));
		bb = single_succ(ENTRY_BLOCK_PTR_FOR_FN(cfun));
	}
	gsi = gsi_start_nondebug_after_labels_bb(bb);
	assign = gimple_build_assign(context, in);
//	gimple_set_location(assign, loc);
	gimple_set_block(assign, DECL_INITIAL(current_function_decl));
	gsi_insert_before(&gsi, assign, GSI_NEW_STMT);
	update_stmt(assign);

	// 3. instrument each BB to track the local context variable
	FOR_EACH_BB_FN(bb, cfun) {
		bb = track_context(bb, context);
	}

	// 4. verify the local context variable against the expected state
	if (EDGE_COUNT(EXIT_BLOCK_PTR_FOR_FN(cfun)->preds)) {
		gcc_assert(single_pred_p(EXIT_BLOCK_PTR_FOR_FN(cfun)));
		gsi = gsi_last_nondebug_bb(single_pred(EXIT_BLOCK_PTR_FOR_FN(cfun)));
		verify_context_before(&gsi, context, out, context_error_decl);
	}

	free_dominance_info(CDI_DOMINATORS);
	free_dominance_info(CDI_POST_DOMINATORS);
	loop_optimizer_finalize();
	return 0;
}

#define PASS_NAME context
#define NO_GATE
#define PROPERTIES_REQUIRED PROP_gimple_leh | PROP_cfg
//#define TODO_FLAGS_START TODO_verify_ssa | TODO_verify_flow | TODO_verify_stmts
#define TODO_FLAGS_FINISH TODO_verify_ssa | TODO_verify_stmts | TODO_dump_func | TODO_verify_flow | TODO_update_ssa
#include "gcc-generate-gimple-pass.h"

#if BUILDING_GCC_VERSION >= 4007 && BUILDING_GCC_VERSION < 8000
static gimple barrier(tree var)
{
	tree input;
#if BUILDING_GCC_VERSION <= 4007
	VEC(tree, gc) *inputs = NULL;
#else
	vec<tree, va_gc> *inputs = NULL;
#endif

	input = build_tree_list(NULL_TREE, build_const_char_string(2, "g"));
	input = chainon(NULL_TREE, build_tree_list(input, var));
#if BUILDING_GCC_VERSION <= 4007
	VEC_safe_push(tree, gc, inputs, input);
#else
	vec_safe_push(inputs, input);
#endif

	return gimple_build_asm_vec("", inputs, NULL, NULL, NULL);
}

static unsigned int fix_esra_execute(void)
{
	basic_block bb;

	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			gimple stmt, barrier_stmt;
			tree lhs, ref;

			stmt = gsi_stmt(gsi);
			if (!gimple_assign_copy_p(stmt))
				continue;

			if (gimple_clobber_p(stmt))
				continue;

			if (TREE_CODE(gimple_assign_rhs1(stmt)) != CONSTRUCTOR)
				continue;

			lhs = gimple_assign_lhs(stmt);
			if (DECL_P(lhs) && DECL_REGISTER(lhs))
				continue;

			ref = get_inner_arg(lhs);
			if (TREE_CODE(ref) == MEM_REF)
				ref = TREE_OPERAND(ref, 0);
			if (TREE_CODE(ref) != SSA_NAME)
				ref = build_unary_op(gimple_location(stmt), ADDR_EXPR, ref, false);
			barrier_stmt = barrier(ref);
//			gimple_asm_set_volatile(as_a_gasm(barrier_stmt), true);
			gimple_set_location(barrier_stmt, gimple_location(stmt));
			gimple_set_block(barrier_stmt, gimple_block(stmt));
			gsi_insert_after(&gsi, barrier_stmt, GSI_NEW_STMT);
		}
	}

	return 0;
}

#define PASS_NAME fix_esra
#define NO_GATE
#define PROPERTIES_REQUIRED PROP_cfg
#define TODO_FLAGS_FINISH TODO_verify_il | TODO_verify_ssa | TODO_verify_stmts | TODO_dump_func | TODO_remove_unused_locals | TODO_update_ssa | TODO_verify_flow | TODO_update_address_taken
#include "gcc-generate-gimple-pass.h"
#endif

static inline void gimple_set_write_once(gimple stmt, bool val_p)
{
	gimple_set_plf(stmt, GF_PLF_2, val_p);
}

static inline bool gimple_write_once_p(gimple stmt)
{
	return gimple_plf(stmt, GF_PLF_2);
}

static inline void gimple_set_read_once(gimple stmt, bool val_p)
{
#if BUILDING_GCC_VERSION < 4007 || (BUILDING_GCC_VERSION == 4007 && BUILDING_GCC_PATCHLEVEL >= 4) || (BUILDING_GCC_VERSION == 4008 && BUILDING_GCC_PATCHLEVEL >= 3)
	stmt->gsbase.pad = val_p;
#elif (BUILDING_GCC_VERSION == 4007 && BUILDING_GCC_PATCHLEVEL < 2)
	gimple_set_in_transaction(stmt, val_p);
#elif (BUILDING_GCC_VERSION == 4007 && BUILDING_GCC_PATCHLEVEL < 4) || (BUILDING_GCC_VERSION == 4008 && BUILDING_GCC_PATCHLEVEL < 3)
	if (val_p)
		*(unsigned int *)&stmt->gsbase |= 0x80000000U;
	else
		*(unsigned int *)&stmt->gsbase &= ~0x80000000U;
#else
	stmt->pad = val_p;
#endif
}

static inline bool gimple_read_once_p(gimple stmt)
{
#if BUILDING_GCC_VERSION < 4007 || (BUILDING_GCC_VERSION == 4007 && BUILDING_GCC_PATCHLEVEL >= 4) || (BUILDING_GCC_VERSION == 4008 && BUILDING_GCC_PATCHLEVEL >= 3)
	return stmt->gsbase.pad == 1;
#elif (BUILDING_GCC_VERSION == 4007 && BUILDING_GCC_PATCHLEVEL < 2)
	return gimple_in_transaction(stmt);
#elif (BUILDING_GCC_VERSION == 4007 && BUILDING_GCC_PATCHLEVEL < 4) || (BUILDING_GCC_VERSION == 4008 && BUILDING_GCC_PATCHLEVEL < 3)
	return *(unsigned int *)&stmt->gsbase & 0x80000000U;
#else
	return stmt->pad == 1;
#endif
}

static void undo_volatile(tree mem)
{
	tree type;
	int quals;

	TREE_THIS_VOLATILE(mem) = 0;
	TREE_SIDE_EFFECTS(mem) = 0;

	type = TREE_TYPE(mem);
	quals = TYPE_QUALS(type) & ~TYPE_QUAL_VOLATILE;
	type = build_qualified_type(type, quals);
	TREE_TYPE(mem) = type;
}

static void redo_volatile(tree mem)
{
	tree type;
	int quals;

	TREE_THIS_VOLATILE(mem) = 1;
	TREE_SIDE_EFFECTS(mem) = 1;

	type = TREE_TYPE(mem);
	quals = TYPE_QUALS(type) | TYPE_QUAL_VOLATILE;
	type = build_qualified_type(type, quals);
	TREE_TYPE(mem) = type;
}

static tree once_do_volatile(tree *tp, int *walk_subtrees, void *data)
{
	tree var = *tp;

	*walk_subtrees = 0;

	switch (TREE_CODE(var)) {
	default:
		debug_tree(var);
		gcc_unreachable();

	case ADDR_EXPR:
	case FIELD_DECL:
	case INTEGER_CST:
	case SSA_NAME:
	case VAR_DECL:
		break;

	case ARRAY_REF:
	case COMPONENT_REF:
	case INDIRECT_REF:
		gcc_assert(TREE_THIS_VOLATILE(var) == TREE_SIDE_EFFECTS(var));
		gcc_assert(!lookup_attribute("once", TYPE_ATTRIBUTES(TREE_TYPE(var))));
		*walk_subtrees = 1;
		break;

#if BUILDING_GCC_VERSION >= 4006
	case MEM_REF:
#if BUILDING_GCC_VERSION >= 4007
		gcc_assert(TREE_THIS_VOLATILE(var) == TREE_SIDE_EFFECTS(var));
#endif
		*walk_subtrees = 1;
		break;
#endif
	}

	return NULL_TREE;
}

static void remove_ssa_names(tree var)
{
	tree ssaname;
	size_t i;

	FOR_EACH_SSA_NAME(i, ssaname, cfun) {
		imm_use_iterator imm_iter;
		gimple stmt;

#if BUILDING_GCC_VERSION >= 4008
		if (virtual_operand_p(ssaname))
			continue;
#endif

		if (SSA_NAME_VAR(ssaname) != var)
			continue;

		FOR_EACH_IMM_USE_STMT(stmt, imm_iter, ssaname) {
			use_operand_p use;

			if (gimple_code(stmt) == GIMPLE_PHI) {
				gimple_stmt_iterator gsi;

				gsi = gsi_for_stmt(stmt);
				remove_phi_node(&gsi, false);
				continue;
			}

			FOR_EACH_IMM_USE_ON_STMT(use, imm_iter) {
				SET_USE(use, var);
			}
			update_stmt(stmt);
		}

		stmt = SSA_NAME_DEF_STMT(ssaname);
		if (SSA_NAME_IS_DEFAULT_DEF(ssaname))
			set_ssa_default_def(cfun, var, NULL_TREE);
		release_ssa_name(ssaname);
		if (!is_gimple_assign(stmt))
			continue;
		gimple_assign_set_lhs(stmt, var);
		update_stmt(stmt);
	}
}

static bool write_once_p(gimple stmt, bool sideeffects)
{
	tree lhs;

	if (!gimple_assign_copy_p(stmt))
		return false;
	if (gimple_clobber_p(stmt))
		return false;

	lhs = gimple_assign_lhs(stmt);

	if (sideeffects && !gimple_has_volatile_ops(stmt))
		return false;

	switch (TREE_CODE(lhs)) {
	default:
		if (lookup_attribute("once", TYPE_ATTRIBUTES(TREE_TYPE(lhs))))
			return true;
		lhs = get_inner_ref(lhs);
		if (lookup_attribute("once", TYPE_ATTRIBUTES(TREE_TYPE(lhs))))
			return true;
#if BUILDING_GCC_VERSION >= 4006
		if (TREE_CODE(lhs) != MEM_REF)
#endif
			return false;
		lhs = TREE_OPERAND(lhs, 1);
		gcc_assert(TREE_CODE(lhs) == INTEGER_CST);
		lhs = TREE_TYPE(lhs);
		gcc_assert(POINTER_TYPE_P(lhs));
		return lookup_attribute("once", TYPE_ATTRIBUTES(TREE_TYPE(lhs))) != NULL_TREE;

	case VAR_DECL:
		if (sideeffects)
			return lookup_attribute("once", TYPE_ATTRIBUTES(TREE_TYPE(lhs))) != NULL_TREE;
		return true;

	case SSA_NAME:
		if (sideeffects)
			return false;
		gcc_assert(SSA_NAME_VAR(lhs) != NULL_TREE);
		return true;
	}
}

static bool write_once(gimple stmt, bool sideeffects)
{
	tree lhs, rhs;

	if (!write_once_p(stmt, sideeffects))
		return false;

	lhs = gimple_assign_lhs(stmt);
	rhs = gimple_assign_rhs1(stmt);

#if BUILDING_GCC_VERSION >= 5000
	gcc_assert(TREE_THIS_VOLATILE(lhs) == TREE_SIDE_EFFECTS(lhs));
#endif
#if BUILDING_GCC_VERSION >= 4006
	gcc_assert(TREE_THIS_VOLATILE(lhs) == sideeffects);
	gcc_assert(TREE_CODE(rhs) != MEM_REF || !lookup_attribute("once", TYPE_ATTRIBUTES(TREE_TYPE(rhs))));
#else
	gcc_unreachable();
#endif

	switch (TREE_CODE(lhs)) {
	default:
		debug_gimple_stmt(stmt);
		debug_tree(lhs);
		gcc_unreachable();

	case COMPONENT_REF:
		if (sideeffects)
			undo_volatile(lhs);
		else
			redo_volatile(lhs);
		walk_tree_without_duplicates(&TREE_OPERAND(lhs, 0), once_do_volatile, sideeffects ? lhs : NULL);
		break;

	case ARRAY_REF:
#if BUILDING_GCC_VERSION >= 4006
	case MEM_REF:
#endif
		if (sideeffects)
			undo_volatile(lhs);
		else
			redo_volatile(lhs);
		walk_tree_without_duplicates(&TREE_OPERAND(lhs, 0), once_do_volatile, sideeffects ? lhs : NULL);
		break;

	case SSA_NAME:
	case VAR_DECL: {
		location_t loc;
		tree var;

		gcc_assert(!sideeffects);

		loc = gimple_location(stmt);
		if (TREE_CODE(lhs) == SSA_NAME) {
			gcc_assert(SSA_NAME_VAR(lhs));
			var = SSA_NAME_VAR(lhs);
		} else
			var = lhs;
		TREE_ADDRESSABLE(var) = 1;
		lhs = build1_loc(loc, ADDR_EXPR, TYPE_POINTER_TO(TREE_TYPE(var)), var);
#if BUILDING_GCC_VERSION >= 4006
		lhs = build2_loc(loc, MEM_REF, TREE_TYPE(TREE_TYPE(lhs)), lhs, build_int_cst(TREE_TYPE(lhs), 0));
#else
		gcc_unreachable();
#endif
		redo_volatile(lhs);
		gimple_assign_set_lhs(stmt, lhs);
		gimple_set_write_once(stmt, sideeffects);
		update_stmt(stmt);

		remove_ssa_names(var);
		return true;
	}
	}

	gimple_set_has_volatile_ops(stmt, !sideeffects);
	gimple_set_write_once(stmt, sideeffects);
	update_stmt(stmt);
	return true;
}

static bool read_once_p(gimple stmt, bool sideeffects)
{
	tree rhs;

	if (!gimple_assign_single_p(stmt))
		return false;
	if (gimple_clobber_p(stmt))
		return false;

	rhs = gimple_assign_rhs1(stmt);

	if (sideeffects && !gimple_has_volatile_ops(stmt))
		return false;

	if (!sideeffects && TREE_CODE(rhs) == SSA_NAME) {
		gcc_assert(SSA_NAME_VAR(rhs) != NULL_TREE);
		return true;
	}

#if BUILDING_GCC_VERSION <= 4007
	if (!sideeffects && TREE_CODE(rhs) == VAR_DECL)
		return true;
#endif

	return lookup_attribute("once", TYPE_ATTRIBUTES(TREE_TYPE(rhs))) != NULL_TREE;
}

static bool read_once(gimple stmt, bool sideeffects)
{
	tree lhs, rhs;

	if (!read_once_p(stmt, sideeffects))
		return false;

	lhs = gimple_assign_lhs(stmt);
	rhs = gimple_assign_rhs1(stmt);

#if BUILDING_GCC_VERSION >= 5000
	gcc_assert(TREE_THIS_VOLATILE(rhs) == TREE_SIDE_EFFECTS(rhs));
#endif
	gcc_assert(TREE_THIS_VOLATILE(rhs) == sideeffects);
#if BUILDING_GCC_VERSION >= 4006
	gcc_assert(TREE_CODE(lhs) != MEM_REF || !lookup_attribute("once", TYPE_ATTRIBUTES(TREE_TYPE(lhs))));
#endif

	switch (TREE_CODE(rhs)) {
	default:
		debug_gimple_stmt(stmt);
		debug_tree(rhs);
		gcc_unreachable();

	case COMPONENT_REF:
		if (sideeffects)
			undo_volatile(rhs);
		else
			redo_volatile(rhs);
		walk_tree_without_duplicates(&TREE_OPERAND(rhs, 0), once_do_volatile, sideeffects ? rhs : NULL);
		break;

	case ARRAY_REF:
	case INDIRECT_REF:
#if BUILDING_GCC_VERSION >= 4006
	case MEM_REF:
#endif
		if (sideeffects)
			undo_volatile(rhs);
		else
			redo_volatile(rhs);
		walk_tree_without_duplicates(&TREE_OPERAND(rhs, 0), once_do_volatile, sideeffects ? rhs : NULL);
		break;

	case PARM_DECL:
	case SSA_NAME:
	case VAR_DECL: {
		tree var;
		location_t loc;

		gcc_assert(!sideeffects);

		loc = gimple_location(stmt);
		if (TREE_CODE(rhs) == SSA_NAME) {
			gcc_assert(SSA_NAME_VAR(rhs));
			var = SSA_NAME_VAR(rhs);
		} else
			var = rhs;
		TREE_ADDRESSABLE(var) = 1;
		rhs = build1_loc(loc, ADDR_EXPR, TYPE_POINTER_TO(TREE_TYPE(var)), var);
#if BUILDING_GCC_VERSION >= 4006
		rhs = build2_loc(loc, MEM_REF, TREE_TYPE(TREE_TYPE(rhs)), rhs, build_int_cst(TREE_TYPE(rhs), 0));
#else
		gcc_unreachable();
#endif
		redo_volatile(rhs);
		gimple_assign_set_rhs1(stmt, rhs);
		gimple_set_read_once(stmt, sideeffects);
		update_stmt(stmt);

		remove_ssa_names(var);
		return true;
	}
	}

	gimple_set_has_volatile_ops(stmt, !sideeffects);
	gimple_set_read_once(stmt, sideeffects);
	update_stmt(stmt);
	return true;
}

static unsigned int once_before_execute(void)
{
	basic_block bb;

	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			gimple stmt;

			stmt = gsi_stmt(gsi);
			gimple_set_write_once(stmt, false);
			gimple_set_read_once(stmt, false);
			write_once(stmt, true) || read_once(stmt, true);
		}
	}

	return 0;
}

#define PASS_NAME once_before
#define NO_GATE
#define PROPERTIES_REQUIRED PROP_cfg
#define TODO_FLAGS_FINISH TODO_verify_il | TODO_verify_ssa | TODO_verify_stmts | TODO_dump_func | TODO_update_ssa | TODO_verify_flow | TODO_update_address_taken
#include "gcc-generate-gimple-pass.h"

static unsigned int once_after_execute(void)
{
	basic_block bb;

	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			gimple stmt;

			stmt = gsi_stmt(gsi);
			gcc_assert(!gimple_read_once_p(stmt) || !gimple_write_once_p(stmt));
			gcc_assert(!gimple_write_once_p(stmt) || write_once(stmt, false));
			gcc_assert(!gimple_read_once_p(stmt) || read_once(stmt, false));
		}
	}

	return 0;
}

static tree ksyms_enum_type;

/* generate ksyms' symbol trees, exported via gcc-ksyms.h */
#define DEFINE(sym, val) __visible const_tree /*GTY(())*/ sym
#define KSYMS_GENERATE_DEFINES
#include "ksyms.def"

static tree ksyms_get_tree_value(tree ksyms, const char *id)
{
	tree field;

	for (field = TYPE_VALUES(ksyms); field; field = TREE_CHAIN(field)) {
		if (id_equal(TREE_PURPOSE(field), id))
			return TREE_VALUE(field);
	}

	error("missing definition of %qs", id);

	return NULL_TREE;
}

/* generate ksyms' symbols population code */
static void ksyms_populate(tree ksyms_type)
{
#define DEFINE(sym, val) sym = ksyms_get_tree_value(ksyms_type, #sym)
#define KSYMS_GENERATE_DEFINES
#include "ksyms.def"
	ksyms_enum_type = ksyms_type;
}

static void finish_type(void *event_data, void *data __unused)
{
	tree type = (tree)event_data;
	tree name, field;

	if (ksyms_enum_type)
		return;

	if (type == NULL_TREE || type == error_mark_node)
		return;

#if BUILDING_GCC_VERSION < 5000
	if (TREE_CODE(type) != RECORD_TYPE)
		return;

	field = TYPE_FIELDS(type);
	if (!field || TREE_CHAIN(field))
		return;

	type = TREE_TYPE(field);
#endif

	if (TREE_CODE(type) != ENUMERAL_TYPE)
		return;

	name = TYPE_NAME(type);
	if (!name || TREE_CODE(name) != IDENTIFIER_NODE || !id_equal(name, "__plugin_ksyms"))
		return;

	/* sanity check type */
	if (TYPE_VALUES(type) == NULL_TREE) {
		error("error parsing %qT", type);
		return;
	}

	for (field = TYPE_VALUES(type); field; field = TREE_CHAIN(field)) {
		const_tree purpose = TREE_PURPOSE(field);
		const_tree value = TREE_VALUE(field);

		gcc_assert(TREE_CODE(purpose) == IDENTIFIER_NODE);
		gcc_assert(TREE_CODE(value) == INTEGER_CST);

//		fprintf(stderr, ">>> added '%s' => %ld\n", IDENTIFIER_POINTER(purpose), tree_to_shwi(value));
	}

//	debug_tree(type);
	ksyms_populate(type);
}

#define PASS_NAME once_after
#define NO_GATE
#define PROPERTIES_REQUIRED PROP_cfg
#define TODO_FLAGS_FINISH TODO_verify_il | TODO_verify_ssa | TODO_verify_stmts | TODO_dump_func | TODO_remove_unused_locals | TODO_update_ssa | TODO_verify_flow | TODO_update_address_taken
#include "gcc-generate-gimple-pass.h"

static unsigned int align_text_bits;
static const char * const align_text_sections[] = {
	".init.text",
	".exit.text",
	".meminit.text",
	".memexit.text",
	".head.text",
	".ref.text",
};

static bool is_align_text_section(const char *name)
{
	size_t i;

	if (!name)
		return false;

	for (i = 0; i < ARRAY_SIZE(align_text_sections); i++)
		if (!strcmp(align_text_sections[i], name))
			return true;

	return false;
}

static void align_init_sections(void *event_data, void *data __unused)
{
	tree fndecl = (tree)event_data;
	location_t loc;
	bool align_me;

	if (fndecl == error_mark_node)
		return;

	if (TREE_CODE(fndecl) != FUNCTION_DECL)
		return;

	gcc_assert(TREE_CODE(TREE_TYPE(fndecl)) == FUNCTION_TYPE);
	loc = DECL_SOURCE_LOCATION(fndecl);

	/* we're done if the alignment is already what we want */
	if (DECL_ALIGN(fndecl) == align_text_bits)
		return;

	align_me = is_align_text_section(get_decl_section_name(fndecl));

	/* skip explicitly aligned functions */
	if (DECL_USER_ALIGN(fndecl)) {
//		if (align_me)
//			error_at(loc, "cannot align %qE, alignment manually set to %qu", fndecl, DECL_ALIGN_UNIT(fndecl));
//		else
//			inform(loc, "would not align %qE, alignment manually set to %qu", fndecl, DECL_ALIGN_UNIT(fndecl));
		return;
	}

	if (!align_me)
		return;

	/* we won't lower alignment */
	if (DECL_ALIGN(fndecl) > align_text_bits) {
		error_at(loc, "not lowering alignment of %qE", fndecl);
		return;
	}

//	inform(loc, "forcing alignment for %qE", fndecl);
	SET_DECL_ALIGN(fndecl, align_text_bits);
//	debug_tree(fndecl);
}

__visible int plugin_init(struct plugin_name_args *plugin_info, struct plugin_gcc_version *version)
{
	const char * const plugin_name = plugin_info->base_name;
	const int argc = plugin_info->argc;
	const struct plugin_argument * const argv = plugin_info->argv;
	int i;
#if BUILDING_GCC_VERSION < 4009
	bool colorize = getenv("GCC_COLORS") ? should_colorize() : false;
#endif
	bool enable_context_tracking = true;

	static const struct ggc_root_tab gt_ggc_r_gt_checker[] = {
		{
			.base = &context_function_decl,
			.nelt = 1,
			.stride = sizeof(context_function_decl),
			.cb = &gt_ggc_mx_tree_node,
			.pchw = &gt_pch_nx_tree_node
		},
		{
			.base = &context_error_decl,
			.nelt = 1,
			.stride = sizeof(context_error_decl),
			.cb = &gt_ggc_mx_tree_node,
			.pchw = &gt_pch_nx_tree_node
		},
		LAST_GGC_ROOT_TAB
	};

	static const struct ggc_root_tab gt_ggc_r_gt_ksyms[] = {
		{
			.base = &ksyms_enum_type,
			.nelt = 1,
			.stride = sizeof(ksyms_enum_type),
			.cb = &gt_ggc_mx_tree_node,
			.pchw = &gt_pch_nx_tree_node
		},
		LAST_GGC_ROOT_TAB
	};

#if BUILDING_GCC_VERSION < 4009
	PASS_INFO(colorize_rearm, "*free_lang_data", 1, PASS_POS_INSERT_AFTER);
#endif

//	PASS_INFO(context, "ssa", 1, PASS_POS_INSERT_AFTER);
//	PASS_INFO(context, "phiprop", 1, PASS_POS_INSERT_AFTER);
	PASS_INFO(context, "*early_warn_uninitialized", 1, PASS_POS_INSERT_AFTER);

#if BUILDING_GCC_VERSION >= 4007 && BUILDING_GCC_VERSION < 8000
	PASS_INFO(fix_esra, "esra", 1, PASS_POS_INSERT_BEFORE);
#endif

	PASS_INFO(once_before, "dse", 0, PASS_POS_INSERT_BEFORE);
	PASS_INFO(once_after, "dse", 0, PASS_POS_INSERT_AFTER);

	if (!plugin_default_version_check(version, &gcc_version)) {
		error_gcc_version(version);
		return 1;
	}

	register_callback(plugin_name, PLUGIN_INFO, NULL, &utilities_plugin_info);

	for (i = 0; i < argc; ++i) {
#if BUILDING_GCC_VERSION < 4009
		if (!strcmp(argv[i].key, "color")) {
			if (!argv[i].value) {
				error(G_("no value supplied for option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
				continue;
			}
			if (!strcmp(argv[i].value, "always"))
				colorize = true;
			else if (!strcmp(argv[i].value, "never"))
				colorize = false;
			else if (!strcmp(argv[i].value, "auto"))
				colorize = should_colorize();
			else
				error(G_("invalid option argument '-fplugin-arg-%s-%s=%s'"), plugin_name, argv[i].key, argv[i].value);
			continue;
		}
#endif
		if (!strcmp(argv[i].key, "nocontext")) {
			enable_context_tracking = false;
			continue;
		}
		if (!strcmp(argv[i].key, "include")) {
			if (!argv[i].value) {
				error(G_("no value supplied for option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
				continue;
			}

			asm_include = xstrdup(argv[i].value);
			continue;
		}
		if (!strcmp(argv[i].key, "align-text-bits")) {
			if (!argv[i].value) {
				error(G_("no value supplied for option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
				continue;
			}

			align_text_bits = atoi(argv[i].value);
			if (align_text_bits == 0 || align_text_bits > 8) {
				error(G_("invalid value supplied for option '-fplugin-arg-%s-%s' (%s)"), plugin_name, argv[i].key, align_text_bits ? "too big" : "too small");
				continue;
			}
			align_text_bits = (1U << align_text_bits) * BITS_PER_UNIT;
			continue;
		}
		error(G_("unknown option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
	}

#if BUILDING_GCC_VERSION < 4009
	if (colorize) {
		// TODO: parse GCC_COLORS as used by gcc 4.9+
		register_callback(plugin_name, PLUGIN_START_UNIT, &colorize_start_unit, NULL);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &colorize_rearm_pass_info);
	}
#endif

#if BUILDING_GCC_VERSION < 8000
	old_internal_error = global_dc->internal_error;
	global_dc->internal_error = emergency_dump_internal_error_function;
#endif

	register_callback(plugin_name, PLUGIN_ATTRIBUTES, register_attributes, NULL);
	register_callback(plugin_name, PLUGIN_START_UNIT, context_start_unit, NULL);
	register_callback(plugin_name, PLUGIN_REGISTER_GGC_ROOTS, NULL, (void *)&gt_ggc_r_gt_checker);
	if (enable_context_tracking)
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &context_pass_info);
	if (asm_include)
		register_callback(plugin_name, PLUGIN_START_UNIT, include_start_unit, NULL);

#if BUILDING_GCC_VERSION >= 4007 && BUILDING_GCC_VERSION < 8000
	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &fix_esra_pass_info);
#endif

	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &once_before_pass_info);
	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &once_after_pass_info);

	register_callback(plugin_name, PLUGIN_FINISH_TYPE, finish_type, NULL);
	register_callback(plugin_name, PLUGIN_REGISTER_GGC_ROOTS, NULL, (void *)&gt_ggc_r_gt_ksyms);

	if (align_text_bits)
		register_callback(plugin_name, PLUGIN_PRE_GENERICIZE, align_init_sections, NULL);

	return 0;
}
